package raytracer

/*-enable-unique*/
import scala.annotation.{unique, transient, uncheckedUnique}

// @transient ListBuffer
import local.collection.mutable.ListBuffer

import scala.actors._
import Actor._

import java.awt.{Image, Dimension}
import java.awt.image.BufferedImage
import java.io.File
import javax.imageio.ImageIO

import scala.Math._
import scala.swing.Swing._

class Raytracer(val screenWidth: Int, val screenHeight: Int) {
  val maxDepth = 5

  object Surfaces {
    val checkerBoard = new Surface {
      override val diffuse = (pos: Vector) => {
        if ((floor(pos.z) + floor(pos.x)) % 2 != 0) new Color(1,1,1) 
        else new Color(0,0,0)
      }
    
      override val specular = (pos: Vector) => {
        new Color(1,1,1)
      }
    
      override val reflect = (pos: Vector) => {
        if ((floor(pos.z) + floor(pos.x)) % 2 != 0) 0.1
	else 0.7
      }
    
      override val roughness: Double = 150
    }

    val shiny = new Surface {
      override val diffuse = (pos: Vector) => { new Color(1,1,1) } 
      override val specular = (pos: Vector) => { new Color(0.5, 0.5, 0.5) }
      override val reflect = (pos: Vector) => 0.6
      override val roughness: Double = 50
    }
  }

  @transient case class PixelBuf(pixels: ListBuffer[(Int, Int, Color)])

  def intersections(ray: Ray, scene: Scene): List[ISect] =
    scene.things
    .map(_.intersect(ray))
    .filter(_ ne null)
    .sortWith((x, y) => (x.dist < y.dist))

  def testRay(ray: Ray, scene: Scene): Double  = {
    val isects = intersections(ray, scene)
      
    isects match {
      case Nil =>  0 
      case x :: xs => { x.dist }
    }
  }

  def traceRay(ray: Ray, scene: Scene, depth: Int): Color = {
    val isects = intersections(ray, scene)
      
    isects match {
      case Nil => Color.background
      case x :: xs => shade(x, scene, depth)
    }
  }

  def getNaturalColor (thing: SceneObject, pos: Vector, norm: Vector, rd: Vector, scene: Scene): Color = {
    var ret = Color.black
    scene.lights.foreach((currentLight: Light) => {
      val ldis = currentLight.position - pos
      val livec = ldis.norm
      val neatIsect = testRay(new Ray(pos, livec), scene)
    
      val isInShadow = !((neatIsect > ldis.mag) || neatIsect == 0)
        
      if (!isInShadow) {
        val illum = livec dot norm
        val lcolor =
          if (illum > 0)
            currentLight.color * illum
          else
            Color.black
        val specular = livec dot (rd.norm)
        val scolor = {
          if (specular > 0) currentLight.color * (pow(specular, thing.surface.roughness))
          else Color.black
        }
        ret = ret + ((lcolor * thing.surface.diffuse(pos)) + (scolor * thing.surface.specular(pos)))
      } 
    }) // end foreach
    ret
  }

  def getReflectionColor(thing: SceneObject, pos: Vector, norm: Vector, rd: Vector, scene: Scene, depth: Int) = {
    traceRay(new Ray(pos, rd), scene, depth + 1) * thing.surface.reflect(pos)
  }

  def shade(isect: ISect, scene: Scene, depth: Int) = {
    val d = isect.ray.dir 
    val pos = (isect.ray.dir * isect.dist) + isect.ray.start
    val normal = isect.thing.normal(pos)
    val reflectDir = d - (normal * (2 * (normal dot d)))

    val ret = Color.defaultColor + getNaturalColor(isect.thing, pos, normal, reflectDir, scene)
    if (depth >= maxDepth) ret + new Color(0.5, 0.5, 0.5)
    else ret + getReflectionColor(isect.thing, pos + (reflectDir * 0.001), normal, reflectDir, scene, depth)
  }

  def recenterX(x: Double) = {
    (x - (screenWidth / 2.0)) / (2.0 * screenWidth)
  }

  def recenterY(y: Double) = {
    -(y - (screenWidth / 2.0)) / (2.0 * screenHeight)
  }

  def getPoint(x: Double, y: Double, camera: Camera) = {
    (camera.forward + ((camera.right * recenterX(x)) + (camera.up * recenterY(y)))).norm
  }

  /** Render `scene` and send resulting image to `dest` actor.
   */
  def render(scene: Scene, dest: Actor) = {
    val image = new BufferedImage(screenWidth, screenHeight, BufferedImage.TYPE_INT_ARGB)
    val parent = self
    val chunkSize = screenHeight / 16
    // create 16 tracer actors
    val tracers = for (chunk <- List.range(0, 16)) yield {
      actor {
        val pixels: ListBuffer[(Int, Int, Color)]@unique = new ListBuffer[(Int, Int, Color)]

        List.range(chunk * chunkSize, (chunk+1) * chunkSize).foreach { y =>
          List.range(0, screenHeight) foreach { x =>
            val color = traceRay(new Ray(scene.camera.pos, getPoint(x, y, scene.camera)), scene, 0)
            pixels += (x, y, color)
          }
        }

        val pixbuf: PixelBuf@unique = PixelBuf(pixels)
        parent ! pixbuf
      }
    }

    var i = 0
    loopWhile (i < tracers.size) {
      i += 1
      react { // receive pixel buffers from tracer actors
        case PixelBuf(pixels) =>
          print("#")
          pixels foreach { case (x, y, c) => image.setRGB(x, y, c.toColor) }
      }
    } andThen {
      println("sending image")
      dest ! (image: @uncheckedUnique)
    }
  }

  /* Objects */
  val objs = List(new Plane(Surfaces.checkerBoard, new Vector(0, 1, 0), 0),
                  new Sphere(Surfaces.shiny, new Vector(0, 1, 0), 1),
                  new Sphere(Surfaces.shiny, new Vector(-1, 0.5, 1.5), 0.5))

  /* Lights */
  val lts = List(new Light(new Vector(-2, 2.5, 0), new Color(0.49, 0.49, 0.49)),
                 new Light(new Vector(1.5, 2.5, 1.5), new Color(0.49, 0.49, 0.49)),
                 new Light(new Vector(1.5, 2.5, -1.5), new Color(0.49, 0.49, 0.49)),
                 new Light(new Vector(0, 3.5, 0), new Color(0.21, 0.21, 0.35)))
  
  val cam = new Camera(new Vector(3,2,4), new Vector(-1, 0.5, 0))
  
  val defaultScene = new Scene(objs, lts, cam)
}

class Vector(val x: Double, val y: Double, val z: Double) {
  def * (n: Double) =
    new Vector(x*n, y*n, z*n)

  def - (v: Vector) =
    new Vector(x - v.x, y - v.y, z - v.z)

  def + (v: Vector) =
    new Vector(x + v.x, y + v.y, z + v.z)

  def dot (v: Vector): Double =
    x * v.x + y * v.y + z * v.z

  def mag =
    sqrt(this dot this)
  
  def norm = {
    val calmag = mag
    val div = if (calmag == 0) Double.PositiveInfinity else { 1 / calmag }
    this * div
  }

  def cross (v: Vector) =
    new Vector((y * v.z) - (z * v.y),
	       (z * v.x) - (x * v.z),
	       (x * v.y) - (y * v.x))
  
  def equals (v: Vector) =
    (x == v.x && y == v.y) && (z == v.z)

  override def toString =
    "(" + x + ", " +y+ ", " +z+ ")"
}

object Color {
  val black = new Color(0, 0, 0)
  val background = black
  val defaultColor = black
}

@transient class Color (val red: Double, val green: Double, val blue: Double) {
  def make(r: Double, g: Double, b: Double) = {
    val c: Color@unique = new Color(r, g, b)
    c
  }

  def * (n: Double) = {
    val c: Color@unique = new Color (n*red, n*green, n*blue)
    c
  }

  def * (color: Color) = {
    val c: Color@unique = new Color (red * color.red, green * color.green, blue * color.blue)
    c
  }

  def + (color: Color) = {
    val c: Color@unique = new Color (red + color.red, green + color.green, blue + color.blue)
    c
  }

  def - (color: Color) = {
    val c: Color@unique = new Color (red - color.red, green - color.green, blue - color.blue)
    c
  }

  override def toString = "Color("+red+", "+green+", "+blue+ ")"
  
  def toColor: Int = {
    def legalize(d: Double): Int = {
      val col = (d * 255).toInt
      if ( col>255 ) 255 else col
    }
    
    val nr = legalize(red)
    val ng = legalize(green)
    val nb = legalize(blue) 
    
    (255 << 24) | (nr << 16) | (ng << 8) | nb;	
  }
}

class Ray(val start: Vector, val dir: Vector)

class ISect(val thing: SceneObject, val ray: Ray, val dist: Double)

abstract class Surface {
  val diffuse: (Vector => Color)
  val specular: (Vector => Color)
  val reflect: (Vector => Double)
  val roughness: Double
}

class Camera (position: Vector, lookAt: Vector) {
  private val down = new Vector(0, -1, 0)
	
  val pos = position
  val forward = (lookAt - position).norm
  val right  = (forward cross down).norm * 1.5
  val up = (forward cross right).norm * 1.5
}

case class Light(val position: Vector, val color: Color)

abstract class SceneObject(val surface: Surface) {
  def intersect(ray: Ray): ISect
  def normal(pos: Vector): Vector
}

class Sphere (sface: Surface, val center: Vector, val radius: Double) extends SceneObject(sface) {

  override def intersect(ray: Ray): ISect = {
    val eo = center - ray.start
    val v: Double  = eo dot ray.dir

    val dist: Double = {
      if (v < 0) 0 
      else {
	val disc = pow(radius,2) - ((eo dot eo) - pow(v,2))
        
	if (disc < 0) 0 
	else (v - sqrt(disc))
      }
    }
    if (dist == 0 || (dist equals Double.NaN)) null
    else new ISect(this, ray, dist)
  }
  
  override def normal(pos: Vector) = {
    (pos - center).norm
  }
}

class Plane(sface: Surface, n: Vector, val offset: Double) extends SceneObject(sface) {
  override def intersect(ray: Ray) = {
    val denom = n dot ray.dir
    if (denom > 0 || (denom equals Double.NaN)) null
    else { 
      val dist = ((n dot ray.start) + offset) / (-denom)
      if (dist equals Double.NaN) null
      else {
        new ISect(this, ray, dist)
      }
    }
  }

  override def normal(pos: Vector) = n
}

class Scene(val things: List[SceneObject], val lights: List[Light], val camera: Camera)

import scala.swing._
import javax.swing.{Icon, ImageIcon}

abstract class SwingActor extends Actor {
  override val scheduler = new SchedulerAdapter {
    def execute(block: => Unit): Unit =
      java.awt.EventQueue.invokeLater(new Runnable() {
        def run() = block
      })
  }
}

object RayTracer extends SimpleSwingApplication {
  val width = 640
  val height = 640

  val images =
    for (_ <- 1 to 4) yield
      new BufferedImage(width/2, height/2, BufferedImage.TYPE_INT_ARGB)

  val labels =
    for (img <- images) yield {
      val lab = new Label
      lab.icon = new ImageIcon(img)
      lab
    }

  def produce(from: Actor) {
    val raytracer = new Raytracer(width, height)
    raytracer.render(raytracer.defaultScene, from)
  }

  val swinger = (new SwingActor {
    def act() {
      var i = 0
      loop {
        i += 1
        react {
          case image: Image =>
            println("\nrecvd image")
            println("currentThread: "+Thread.currentThread)
            labels(i-1).icon = new ImageIcon(image.getScaledInstance(width/2, height/2, Image.SCALE_FAST))
            labels(i-1).repaint
          case time: Long =>
            i -= 1
            timeLabel.text = "Time: "+time+" ms"
        }
      }
    }
  }).start()

  val timeLabel = new Label("Time")
  timeLabel.size = ((width, width/10): Dimension)
  timeLabel.font = new java.awt.Font(java.awt.Font.SANS_SERIF, java.awt.Font.BOLD, 32)

  actor {
    self.trapExit = true
    val start = System.currentTimeMillis
    for (_ <- 1 to 4)
      link { produce(swinger) }
    for (_ <- 1 to 4)
      receive { case _ => }
    val end = System.currentTimeMillis
    swinger ! (end-start)
  }

  def top = new MainFrame {
    size = ((600, 600): Dimension)
    title = "Raytracer"
    contents = {
      val grid = new GridPanel(2, 2) {
        for (lab <- labels)
          contents += lab
        border = Swing.EmptyBorder(10, 10, 10, 10)
      }
      val panel = new BoxPanel(Orientation.Vertical) {
        contents.append(timeLabel, grid)
      }
      panel
    }
  }
}
